#ifndef _H_CX_COLLECTIONS_BUFFER
#define _H_CX_COLLECTIONS_BUFFER

#include <cx/libc.hpp>
#include <cx/stddef.hpp>
#include <cx/stl.hpp>

template <typename T>
concept Buffer = requires(T t, uint8_t byte, uint8_t const* byte_ptr, size_t byte_count) {
    { t.buf_get_raw_ptr() } -> stl::same_as<uint8_t const*>;
    { t.buf_get_raw_mut_ptr() } -> stl::same_as<uint8_t*>;
    { t.buf_get_size() } -> stl::same_as<size_t>;
    { t.buf_push_byte(byte) } -> stl::same_as<void>;
    { t.buf_append_bytes(byte_ptr, byte_count) } -> stl::same_as<void>;
};

class BufferContainer {
private:
    uint8_t* inner;
    size_t len;
    size_t size;

public:
    BufferContainer() {
        this->inner = (uint8_t*)libc::malloc(1024);
        this->size = 1024;
    }

    ~BufferContainer() {
        libc::free(this->inner);
    }

    void resize(size_t bytes) {
        this->inner = (uint8_t*)libc::realloc(this->inner, bytes);
    }

    // IMPL: Buffer
    uint8_t const* buf_get_raw_ptr() {
        return (uint8_t const*)this->inner;
    }

    // IMPL: Buffer
    size_t buf_get_size() {
        return this->size;
    }

    // IMPL: Buffer
    void buf_push_byte(uint8_t byte) {
        if (this->size <= this->len) {
            this->resize(this->size * 2);
        }
        this->inner[this->len++] = byte;
    }

    // IMPL: Buffer
    void buf_append_bytes(uint8_t const* byte_ptr, size_t byte_count) {
        for (size_t i = 0; i < byte_count; i++) {
            this->buf_push_byte(byte_ptr[i]);
        }
    }
};

#endif // !_H_CX_COLLECTIONS_BUFFER