#include <string.h>
#include <cx/string.hpp>

String::String(const char* str) {
    this->_chars = Vec<char>::from_ptr((char*)str, strlen(str));
}

size_t String::len() {
    return this->_chars.len();
}

void String::clear() {
    this->_chars.clear();
}

Vec<char>& String::chars() {
    return this->_chars;
}

String String::fmt_to_str() {
    return this->clone();
}

char* const String::as_cstr() {
    return this->_chars.to_ptr();
}

void String::push(char c) {
    this->_chars.push(c);
}

void String::push_str(String& str) {
    Vec<char> chrs = str.chars();

    for (size_t i = 0; i < chrs.len(); i++) {
        // SAFETY: Always Valid
        this->push(*chrs.nth(i).unwrap());
    }
}

void String::push_str(const char* str) {
    for (size_t i = 0; i < strlen(str); i++) {
        this->push(str[i]);
    }
}

// IMPL: Clone
String String::clone() {
    String s = String();
    s._chars = this->_chars.clone();
    return s;
}

// IMPL: Buffer
uint8_t const* String::buf_get_raw_ptr() {
    return (uint8_t const*)this->_chars.to_ptr();
}

// IMPL: Buffer
uint8_t* String::buf_get_raw_mut_ptr() {
    return (uint8_t*)this->_chars.to_ptr();
}

// IMPL: Buffer
size_t String::buf_get_size() {
    return this->len();
}

void String::buf_push_byte(uint8_t byte) {
    this->push((char)byte);
}

void String::buf_append_bytes(uint8_t const* byte_ptr, size_t byte_count) {
    for (size_t i = 0; i < byte_count; i++) {
        this->buf_push_byte(byte_ptr[i]);
    }
}